import Compression from "compression";
import CookieParser from "cookie-parser";
import cors from "cors";
import crypto from "crypto";
import Express from "express";
import helmet from "helmet";
import * as Http from "http";
import { createProxyMiddleware } from "http-proxy-middleware";
import internalIp from "internal-ip";
import Path from "path";
import { URL } from "url";
import { getPort, getPublicPath } from "./utils/getConfig.js";
import getLogger from "./utils/logger.js";
import checkPositivePath from "./utils/checkPositivePath.js";
const protocalPortMap = {
    "ftp:": 21,
    "http:": 80,
    "https:": 443,
    "ws:": 80,
    "wss:": 443,
};
const logger = getLogger();
const log = logger.getLogger("server/index");
class WebServer {
    state = false;
    options;
    routers;
    name = "WebServer";
    server;
    app;
    constructor(options) {
        const defaultPort = getPort();
        if ((!options || !options.port) && !defaultPort) {
            throw new Error("no port specified");
        }
        const defautlOptions = {
            notSSR: false,
            port: getPort(),
        };
        this.options = {
            ...defautlOptions,
            ...options,
        };
        this.setupApp();
        this.setHealthCheck();
        this.errorEventHandler();
        this.createServer();
    }
    checkServerStarted() {
        if (this.state) {
            log.warn("server started, can not bootstrap or set router");
        }
        return this.state;
    }
    setupApp() {
        this.app = Express();
    }
    createServer() {
        this.server = Http.createServer(this.app);
        this.server.on("error", e => {
            let emsg = "server start error: " + e.message;
            if (e.code === 'EADDRINUSE') {
                emsg = `server cannot start, address in use: ${this.options.hostname || '127.0.0.1'}:${this.options.port}`;
                console.error(emsg);
            }
            log.error(emsg);
            this.close();
            throw new Error(emsg);
        });
        this.server.on("listening", () => {
            this.state = true;
            console.log("server start successful, listening at port: " + this.options.port);
            if (typeof this.options.onListening === "function") {
                this.options.onListening(this.server);
            }
        });
    }
    getInteralStatic(staticOption = {}) {
        const staticRootPath = getPublicPath();
        log.info("setStatic recieve staticRootPath: ", staticRootPath);
        const path = staticRootPath.replace(/\/$/, "");
        const directory = Path.resolve(staticRootPath.replace(/^\//, ""));
        log.info("static path : ", path);
        log.info("static directory : ", directory);
        return {
            path: [path],
            directory: directory,
            staticOption,
        };
    }
    setCloseHandle() {
        process.once("SIGINT", () => {
            this.close();
        });
    }
    setHealthCheck() {
        this.app.get("/status", (req, res) => {
            res.send("OK");
        });
    }
    errorEventHandler() {
        process.on("uncaughtException", (error) => {
            log.error(this.name, "errorEventHandler", "UNCAUGHT_EXCEPTION", "!!!未处理的严重异常.被process.uncaughtException捕获!!!", error);
        });
    }
    async listen(port, hostname) {
        return new Promise((resolve, reject) => {
            (async () => {
                try {
                    if (hostname === "local-ip") {
                        hostname = await internalIp.v4() || await internalIp.v6() || "0.0.0.0";
                    }
                    else if (hostname === "local-ipv4") {
                        hostname = await internalIp.v4() || "0.0.0.0";
                    }
                    else if (hostname === 'local-ipv6') {
                        hostname = await internalIp.v6() || '::';
                    }
                    this.server.listen({
                        port: port,
                        host: hostname,
                    });
                    resolve("");
                }
                catch (e) {
                    reject(e);
                }
            })();
        });
    }
    setRouter() {
        if (this.routers) {
            this.routers.forEach(router => this.app.use(router));
        }
    }
    setMiddleware() {
        const features = {
            notSSR: () => {
                this.setNotSSRQueryFeature();
            },
            helmet: () => {
                this.setupHelmetFeature();
            },
            cors: () => {
                this.setupCorsFeature();
            },
            access: () => {
                this.setupAccessFeature();
            },
            compress: () => {
                if (this.options.compress) {
                    this.setupCompressFeature();
                }
            },
            cookieParser: () => {
                if (this.options.cookieParser) {
                    this.setupCookieParserFeature();
                }
            },
            bodyParser: () => {
                if (this.options.bodyParser) {
                    this.setupBodyParserFeature();
                }
            },
            proxy: () => {
                if (this.options.proxy) {
                    this.setupProxyFeature();
                }
            },
            static: () => {
                this.setupStaticFeature();
            },
            middleware: () => {
                if (this.options.middleware) {
                    this.setupMiddleware();
                }
            },
            headers: () => {
                if (this.options.headers) {
                    this.setupHeadersFeature();
                }
            },
        };
        const runnableFeatures = [];
        runnableFeatures.push("access");
        if (this.options.compress) {
            runnableFeatures.push("compress");
        }
        if (this.options.headers) {
            runnableFeatures.push("headers");
        }
        if (this.options.cookieParser) {
            runnableFeatures.push("cookieParser");
        }
        if (this.options.proxy) {
            runnableFeatures.push("proxy");
        }
        if (this.options.bodyParser) {
            runnableFeatures.push("bodyParser");
        }
        if (this.options.secureHeaderOptions !== false) {
            runnableFeatures.push("helmet");
        }
        runnableFeatures.push("static");
        if (this.options.middleware) {
            runnableFeatures.push("middleware");
        }
        if (this.options.corsOptions) {
            runnableFeatures.push("cors");
        }
        if (this.options.notSSR) {
            runnableFeatures.push("notSSR");
        }
        runnableFeatures.forEach((feature) => {
            features[feature]();
        });
    }
    setNotSSRQueryFeature() {
        this.app.use((req, res, next) => {
            req.query._notssr = "1";
            next();
        });
    }
    setupAccessFeature() {
        if (this.options.access) {
            this.app.use(this.options.access);
        }
        else {
            this.app.use(logger.connectLogger(logger.getLogger("http"), { level: "auto", }));
        }
    }
    setupMiddleware() {
        const mws = Object.prototype.toString.call(this.options.middleware) === "[object Array]"
            ? this.options.middleware : [this.options.middleware];
        mws.forEach(mw => {
            this.app.use(mw);
        });
    }
    setupCorsFeature() {
        if (Array.isArray(this.options.corsOptions)) {
            this.options.corsOptions.forEach((config) => {
                this.app.use(config.path, cors(config.corsOptions));
            });
        }
        else {
            this.app.use(cors(this.options.corsOptions));
        }
    }
    setupHelmetFeature() {
        const options = {
            ...this.options?.secureHeaderOptions,
            contentSecurityPolicy: {
                ...this.options.secureHeaderOptions?.contentSecurityPolicy,
                directives: {
                    ...this.options.secureHeaderOptions?.contentSecurityPolicy?.directives,
                    scriptSrc: [
                        "'self'",
                        (req, res) => {
                            const fsResponsiveNonce = crypto.randomBytes(16).toString('base64');
                            res.locals.fsResponsiveNonce = fsResponsiveNonce;
                            return `'nonce-${fsResponsiveNonce}'`;
                        },
                        (req, res) => {
                            const csDataNonce = crypto.randomBytes(16).toString('base64');
                            res.locals.csDataNonce = csDataNonce;
                            return `'nonce-${res.locals.csDataNonce}'`;
                        },
                    ].concat(this.options.secureHeaderOptions?.contentSecurityPolicy?.directives?.scriptSrc || []),
                },
            }
        };
        this.app.use(helmet(options));
    }
    setupCompressFeature() {
        this.app.use(Compression());
    }
    setupCookieParserFeature() {
        this.app.use(CookieParser());
    }
    changeProxyReqReferer(proxyReq, req, res) {
        const refererURL = new URL(req.header("Referer"));
        const hostStr = proxyReq.getHeader("Host");
        refererURL.host = hostStr;
        if (hostStr.indexOf(":") === -1) {
            refererURL.port = protocalPortMap[refererURL.protocol];
        }
        proxyReq.setHeader("Referer", refererURL.href);
    }
    getProxyMW(config) {
        const opt = {
            changeOrigin: true,
            changeOriginReferer: true,
            pathRewrite: {
                "^/proxy": "",
            },
            onError: (err, req, res, target) => {
                res.writeHead(500, {
                    "Content-Type": "text/plain",
                });
                res.end("proxying request error: " + err.message);
            },
            ...config,
        };
        if (opt.changeOrigin === false) {
            opt.changeOriginReferer = false;
        }
        if (opt.changeOriginReferer !== false) {
            opt.onProxyReq = this.changeProxyReqReferer;
        }
        return createProxyMiddleware(opt);
    }
    setupProxyFeature() {
        const proxyConfig = this.options.proxy;
        if (typeof proxyConfig === "string") {
            this.app.use("/proxy", this.getProxyMW({
                target: proxyConfig,
            }));
        }
        else if (Array.isArray(proxyConfig)) {
            proxyConfig.forEach((proxy) => {
                this.app.use("/proxy" + proxy.path, this.getProxyMW({
                    pathRewrite: undefined,
                    ...proxy.config,
                }));
            });
        }
        else if (proxyConfig) {
            this.app.use("/proxy", this.getProxyMW({
                router: proxyConfig,
            }));
        }
    }
    setupBodyParserFeature() {
        const bodyParserOpt = {
            limit: 10 * 1024 * 1024,
        };
        let rawOpt = {
            ...bodyParserOpt,
        };
        let jsonOpt = {
            ...bodyParserOpt,
        };
        let textOpt = {
            ...bodyParserOpt,
        };
        let urlencodedOpt = {
            ...bodyParserOpt,
            extended: true,
        };
        if (Object.prototype.toString.call(this.options.bodyParser) === "[object Object]") {
            if (Object.prototype.toString.call(this.options.bodyParser.raw) === "[object Object]") {
                rawOpt = this.options.bodyParser.raw;
            }
            if (Object.prototype.toString.call(this.options.bodyParser.json) === "[object Object]") {
                jsonOpt = this.options.bodyParser.json;
            }
            if (Object.prototype.toString.call(this.options.bodyParser.text) === "[object Object]") {
                textOpt = this.options.bodyParser.text;
            }
            if (Object.prototype.toString.call(this.options.bodyParser.urlencoded) === "[object Object]") {
                urlencodedOpt = this.options.bodyParser.urlencoded;
            }
        }
        if (typeof this.options.bodyParser === "boolean") {
            this.app.use(Express.raw(rawOpt));
            this.app.use(Express.json(jsonOpt));
            this.app.use(Express.text(textOpt));
            this.app.use(Express.urlencoded(urlencodedOpt));
        }
        else {
            if (this.options.bodyParser.raw) {
                this.app.use(Express.raw(rawOpt));
            }
            if (this.options.bodyParser.json) {
                this.app.use(Express.json(jsonOpt));
            }
            if (this.options.bodyParser.text) {
                this.app.use(Express.text(textOpt));
            }
            if (this.options.bodyParser.urlencoded) {
                this.app.use(Express.urlencoded(urlencodedOpt));
            }
        }
    }
    setupHeadersFeature() {
        this.app.all('*', this.setContentHeaders.bind(this));
    }
    setContentHeaders(req, res, next) {
        for (const name in this.options.headers) {
            res.setHeader(name, this.options.headers[name]);
        }
        next();
    }
    setupStaticFeature() {
        if (Object.prototype.toString.call(this.options.static) !== "[object Array]") {
            this.options.static = [];
        }
        const internalStatic = this.options.static[0];
        let staticOption = {};
        if (internalStatic && internalStatic.staticOption && internalStatic.isInternal) {
            staticOption = internalStatic.staticOption;
        }
        if (!DEV_STATIC_HR_SERVE) {
            this.options.static.splice(0, 0, this.getInteralStatic(staticOption));
        }
        this.options.static.forEach(staticOption => {
            staticOption.path.forEach(path => {
                if (!checkPositivePath(path)) {
                    this.app.use(path, Express.static(staticOption.directory, staticOption.staticOptions));
                }
            });
        });
    }
    ready() {
        if (IS_DEBUG_MODE) {
            const proxConfig = DEV_PROXY_CONFIG;
            if (proxConfig && Array.isArray(proxConfig)) {
                proxConfig.forEach((proxy) => {
                    this.app.use(proxy.path, this.getProxyMW({
                        pathRewrite: undefined,
                        ...proxy.config,
                    }));
                });
            }
        }
        this.setMiddleware();
        this.setRouter();
    }
    async bootstrapAsync() {
        if (!this.checkServerStarted()) {
            await this.bootstrapAsyncBefore();
            this.ready();
            this.setCloseHandle();
            await this.listen(this.options.port, this.options.hostname);
            log.info(this.name, "bootstrapAsync", "port", this.options.port);
        }
        return this;
    }
    useRouter(router) {
        if (!this.checkServerStarted()) {
            if (!this.routers) {
                this.routers = [];
            }
            this.routers.push(router);
        }
        return this;
    }
    close(cb) {
        this.server.close(() => {
            this.state = false;
            log.warn(this.name, "close", "port", this.options.port);
            if (typeof this.options.onServerClosed === "function") {
                this.options.onServerClosed();
            }
            if (typeof cb === "function") {
                cb();
            }
        });
    }
    async bootstrapAsyncBefore() {
    }
}
export default WebServer;
